@@ -4,6 +4,7 @@ import os
 import re
 import json
 import shutil
+from dataclasses import dataclass, field
 from pathlib import Path
 from typing import List, Tuple, Iterable, Optional
 
@@ -28,7 +29,7 @@ def detect_spaces_per_level(lines: Iterable[str]) -> int:
     indents = []
     for raw in lines:
         normalized = normalize_line_for_parsing(raw)
-        if not normalized.strip(): 
+        if not normalized.strip():
             continue
         leading = len(normalized) - len(normalized.lstrip(" "))
         if leading > 0:
@@ -37,7 +38,7 @@ def detect_spaces_per_level(lines: Iterable[str]) -> int:
         return 4
     base = min(indents)
     if base >= 8:
-        for d in (4,2):
+        for d in (4, 2):
             if all((x % d == 0) for x in indents):
                 return d
     return base or 4
@@ -45,7 +46,7 @@ def detect_spaces_per_level(lines: Iterable[str]) -> int:
 def guess_is_file(name: str) -> bool:
     if not name:
         return False
-    if name.endswith("/"): 
+    if name.endswith("/"):
         return False
     simple = name.rstrip("/").strip()
     if simple.startswith(".") and simple.count(".") == 1:
@@ -171,59 +172,115 @@ def create_from_components(
                 if verbose: print(f"Created directory: {target}")
     return created, skipped
 
-# ---- scan directory -> tree text ----
+# ---- scan directory -> hierarchical tree, then render consistently ----
 
-def _format_ascii_tree(lines: List[Tuple[int, str, bool]]) -> List[str]:
-    # Given list of (depth, name, is_file), build ascii-tree style lines
-    out = []
-    last_at_depth = {}
-    total = len(lines)
-    for idx, (depth, name, is_file) in enumerate(lines):
-        # determine connector char: last child vs not
-        # find if next item at same or shallower depth exists to consider 'has sibling'
-        has_sibling = False
-        for j in range(idx+1, total):
-            if lines[j][0] < depth:
-                break
-            if lines[j][0] == depth:
-                has_sibling = True
-                break
-        prefix = ""
-        if depth > 0:
-            parts = []
-            for d in range(depth):
-                parts.append("│   " if any(l[0] > d and l[0] >= d+1 for l in lines) else "    ")
-            prefix = "".join(parts)
-            connector = "├── " if has_sibling else "└── "
-        else:
-            connector = ""
-        out.append(f"{prefix}{connector}{name}")
+@dataclass
+class Node:
+    name: str
+    is_file: bool
+    children: list["Node"] = field(default_factory=list)
+
+def _safe_is_file(p: Path) -> bool:
+    try:
+        return p.is_file()
+    except OSError:
+        return True
+
+def build_tree(root: Path, follow_symlinks: bool=False, max_depth: Optional[int]=None) -> Node:
+    """
+    Build a directory tree as nested Node objects.
+    Directories are listed before files; items are sorted case-insensitively.
+    Symlinks are treated as files unless follow_symlinks=True. Cycles are avoided.
+    """
+    root = root.resolve()
+    visited: set[Tuple[int, int]] = set()
+
+    def _mark_visited(p: Path):
+        try:
+            st = p.stat()
+            visited.add((st.st_dev, st.st_ino))
+        except OSError:
+            pass
+
+    def _is_visited(p: Path) -> bool:
+        try:
+            st = p.stat()
+            return (st.st_dev, st.st_ino) in visited
+        except OSError:
+            return True
+
+    _mark_visited(root)
+
+    def _build(path: Path, depth: int) -> Node:
+        node = Node(path.name, _safe_is_file(path))
+        if node.is_file:
+            return node
+        if max_depth is not None and depth >= max_depth:
+            return node
+        try:
+            entries = sorted(
+                list(path.iterdir()),
+                key=lambda x: (_safe_is_file(x), x.name.lower())
+            )
+        except (PermissionError, OSError):
+            return node
+
+        for entry in entries:
+            if entry.is_symlink():
+                if not follow_symlinks:
+                    node.children.append(Node(entry.name, True))
+                    continue
+                if _is_visited(entry):
+                    node.children.append(Node(entry.name, True))
+                    continue
+                _mark_visited(entry)
+
+            child = _build(entry, depth + 1)
+            node.children.append(child)
+        return node
+
+    return _build(root, 0)
+
+def _flatten(node: Node, depth: int = 0, out: Optional[List[Tuple[int, str, bool]]] = None) -> List[Tuple[int,str,bool]]:
+    if out is None:
+        out = []
+    out.append((depth, node.name, node.is_file))
+    if not node.is_file:
+        for child in node.children:
+            _flatten(child, depth+1, out)
     return out
 
-def scan_dir_to_entries(root: Path, follow_symlinks: bool=False, max_depth: Optional[int]=None) -> List[Tuple[int,str,bool]]:
-    entries: List[Tuple[int,str,bool]] = []
-    root = root.resolve()
-    base_parts = len(root.parts)
-    for dirpath, dirnames, filenames in os.walk(root, followlinks=follow_symlinks):
-        path = Path(dirpath)
-        depth = len(path.parts) - base_parts
-        rel = path.name if depth>=0 else str(path)
-        # include directory itself (except root)
-        if depth >= 0:
-            entries.append((depth, path.name, False))
-        # directories (sorted)
-        for d in sorted(dirnames):
-            entries.append((depth+1, d, False))
-        for f in sorted(filenames):
-            entries.append((depth+1, f, True))
-        if max_depth is not None and depth >= max_depth:
-            # prune from walking deeper
-            dirnames[:] = []
-    return entries
+def _render_ascii_from_node(root: Node) -> List[str]:
+    lines: List[str] = [root.name]
 
-def render_entries_as_indented(entries: List[Tuple[int,str,bool]], spaces:int=4, use_tabs:bool=False, emoji:Optional[Tuple[str,str]]=None) -> List[str]:
-    out = []
-    for depth, name, is_file in entries:
+    def rec(children: List[Node], prefix: str) -> None:
+        for i, child in enumerate(children):
+            last = (i == len(children) - 1)
+            connector = "└── " if last else "├── "
+            lines.append(prefix + connector + child.name)
+            if not child.is_file and child.children:
+                extension = "    " if last else "│   "
+                rec(child.children, prefix + extension)
+
+    if not root.is_file and root.children:
+        rec(root.children, "")
+    return lines
+
+# Public API (kept names but now Node-powered under the hood)
+
+def scan_dir_to_entries(root: Path, follow_symlinks: bool=False, max_depth: Optional[int]=None):
+    """
+    Backwards-compatible name, but returns a Node tree now.
+    """
+    return build_tree(root, follow_symlinks=follow_symlinks, max_depth=max_depth)
+
+def render_entries_as_indented(root_node, spaces:int=4, use_tabs:bool=False, emoji:Optional[Tuple[str,str]]=None) -> List[str]:
+    """
+    Render from Node into an indented list (tabs/spaces, optional emojis).
+    """
+    flat = _flatten(root_node)
+    out: List[str] = []
+    for depth, name, is_file in flat:
         indent = ("\t" * depth) if use_tabs else (" " * (spaces * depth))
         if emoji:
             folder_emoji, file_emoji = emoji
@@ -233,16 +290,17 @@ def render_entries_as_indented(entries: List[Tuple[int,str,bool]], spaces:int=4,
             out.append(f"{indent}{name}")
     return out
 
-def render_entries_as_ascii(entries: List[Tuple[int,str,bool]]) -> List[str]:
-    # Convert entries into (depth,name,is_file) list for ascii helper
-    # The ascii helper expects ordered depth list. We will rebuild a simpler ordering:
-    flat = []
-    # We want to emulate 'tree' order: directory, then its children.
-    # entries from scan_dir_to_entries already roughly in that order.
-    for depth,name,is_file in entries:
-        flat.append((depth, name, is_file))
-    return _format_ascii_tree(flat)
+def render_entries_as_ascii(root_node) -> List[str]:
+    """
+    Render from Node into ASCII tree (├──, │, └──), perfectly aligned.
+    """
+    return _render_ascii_from_node(root_node)
 
-def entries_to_json(entries: List[Tuple[int,str,bool]]) -> str:
-    arr = [{"depth": d, "path": name, "is_file": is_file} for d,name,is_file in entries]
+def entries_to_json(root_node) -> str:
+    """
+    JSON dump (depth + name + is_file) derived from the same Node tree
+    used by all other renderers.
+    """
+    flat = _flatten(root_node)
+    arr = [{"depth": d, "path": n, "is_file": is_f} for d, n, is_f in flat]
     return json.dumps(arr, indent=2, ensure_ascii=False)
